/*
 * granular uci access control
 *
 * Copyright (C) 2019 IOPSYS
 *
 * Author: Jakob Olsson <jakob.olsson@iopsys.eu>
 * Author: Matija Amidzic <matija.amidzic@sartura.hr>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <libubox/avl.h>
#include <libubox/blobmsg_json.h>

#define GRANULAR_MAX_MATCH_OPTIONS 8	 	// maximum number of options by which to match the section (with uci lookup)
#define GRANULAR_MAX_ALLOWED_OPTIONS 16  	// maximum size of list of allowed options
#define GRANULAR_MAX_MATCH_DEFINITIONS 8 	// maximum number of match definitions per config
#define GRANULAR_MAX_STRING_LENGTH 64
#define GRANULAR_MAX_MATCHED_DEFINITIONS 32 // maximum number of matched definitions to parse when giving access

struct uci_granular_query_s {
    const char *config_name;
    const char *username;
    const char *section_name;
    const char *section_type;
};

/* high level of access has higher index */
enum option_access {
	RPC_OA_NONE,
	RPC_OA_OPTION,
	RPC_OA_SECTION,
	RPC_OA_FULL,
};

enum config_access {
	RPC_CA_ALLOWED, 	// No restrictions at all in place (no uci_granular at all)
	RPC_CA_RESTRICTED,	// Restrictions of some sort are in place
};

struct match_option_s {
	char name[GRANULAR_MAX_STRING_LENGTH];
	char value[GRANULAR_MAX_STRING_LENGTH];
};

struct match_definition_s {
	char match_section_type[GRANULAR_MAX_STRING_LENGTH];
	char match_section_name[GRANULAR_MAX_STRING_LENGTH];

	struct match_option_s match_options[GRANULAR_MAX_MATCH_OPTIONS];
	int match_options_count;

	char **allowed_options;
	int allowed_options_count;
};

struct uci_granular_ctx_s {
	struct match_definition_s *match_definitions[GRANULAR_MAX_MATCHED_DEFINITIONS];

	/* Restriction placed on the sought after config file
	 * -1: Free access, no match/option pair found for configuration file
	 * 0: No access (restriction, but no options listed)
	 * 0<: Access restricted by match/option pair
	 */
	int match_definitions_count;

	enum config_access state;
};



int rpc_uci_granular_init();
int rpc_uci_granular_setup(struct rpc_session *ses, struct blob_attr *acl_top, const char *username);
void rpc_uci_granular_get_allowed(struct rpc_session *ses, struct uci_granular_query_s *query, struct uci_granular_ctx_s *granular_ctx);
void uci_granular_exit(struct rpc_session *ses);
int uci_granular_parse_match_definition(struct blob_attr *acl_match_definition, struct match_definition_s *match_definition);
int rpc_uci_granular_destroy(void);
