#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include <cmocka.h>
#include <rpcd/uci.h>
#include <json-c/json.h>
#include <rpcd/session.h>
#include <rpcd/uci_granular.h>
#include <libubox/blobmsg.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <libubox/blobmsg_json.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <signal.h>
#include <sys/stat.h>

#include <rpcd/plugin.h>
#include <rpcd/exec.h>
#include <json-editor.h>

#define USER "user"
#define PASSWORD "user"
#define CFG "test"

int
rpc_uci_get(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg);
int rpc_uci_set(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg);
int rpc_uci_revert_commit(struct ubus_context *ctx, struct blob_attr *msg,
		bool commit);
int rpc_uci_add(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg);
int rpc_uci_delete(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg);
int rpc_uci_rename(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg);
void rpc_uci_trigger_event(struct ubus_context *ctx, const char *config);

struct test_env {
	struct rpc_session *ses;
	struct uci_context *cursor;
};

static struct blob_attr *ba;

static void reset_ba()
{
	if (ba) {
		free(ba);
		ba = NULL;
	}
}

/**
 * override libubus function
 */
int ubus_send_reply(struct ubus_context *ctx, struct ubus_request_data *req,
                struct blob_attr *msg)
{
	int len;
	char *str;

	str = blobmsg_format_json(msg, true);

	printf("here comes the reponse %s\n", str);

	reset_ba();
	len = blob_raw_len(msg);
	ba = (struct blob_attr *) calloc(1, sizeof(struct blob_attr) + len);
	memcpy(ba, msg, len);

	free(str);
	return 0;
}

void rpc_uci_trigger_event(struct ubus_context *ctx, const char *config)
{
	return;
}

static int cp(const char *to, const char *from)
{
    int fd_to, fd_from;
    char buf[4096];
    ssize_t nread;
    int saved_errno;

    fd_from = open(from, O_RDONLY);
    if (fd_from < 0)
        return -1;

    fd_to = open(to, O_WRONLY | O_CREAT | O_EXCL, 0666);
    if (fd_to < 0)
        goto out_error;

    while (nread = read(fd_from, buf, sizeof buf), nread > 0) {
        char *out_ptr = buf;
        ssize_t nwritten;

        do {
            nwritten = write(fd_to, out_ptr, nread);

            if (nwritten >= 0) {
                nread -= nwritten;
                out_ptr += nwritten;
            } else if (errno != EINTR) {
                goto out_error;
            }
        } while (nread > 0);
    }

    if (nread == 0) {
        if (close(fd_to) < 0) {
            fd_to = -1;
            goto out_error;
        }
        close(fd_from);

        /* Success! */
        return 0;
    }

  out_error:
    saved_errno = errno;

    close(fd_from);
    if (fd_to >= 0)
        close(fd_to);

    errno = saved_errno;
    return -1;
}

static void destroy_env(struct test_env *e)
{
	uci_free_context(e->cursor);

	rpc_session_destroy(e->ses);
	free(e);
}

static struct test_env *create_new_session(char *user)
{
	struct test_env *e;
	struct rpc_session *ses = NULL;
	struct uci_section *login;
	int timeout = 300;
	int rv;

	e = calloc(1, sizeof(struct test_env));
	if (!e)
		return NULL;

	e->cursor = uci_alloc_context();
	if (!e->cursor)
		goto out;

	login = rpc_login_test_login(e->cursor, user, user, NULL);
	if (!login)
		goto out_uci;

	if (!rpc_user_login_enabled(user))
		goto out_uci;

	ses = rpc_session_create(timeout);
	if (!ses)
		goto out_uci;

	rv = rpc_uci_granular_init(ses);
	if (rv)
		goto out_session;
	rpc_login_setup_acls(ses, login);
	e->ses = ses;

	return e;

out_session:
	rpc_session_destroy(e->ses);
out_uci:
	uci_free_context(e->cursor);
out:
	free(e);
	return NULL;
}

static void test_settings_name_access(void **state) {
	struct test_env *e = (struct test_env *) *state;

	struct uci_granular_ctx_s *granular_ctx = rpc_uci_load_granular(USER, CFG, "settings", "settings", e->ses->id);

	assert_int_equal(granular_ctx->match_definitions_count, 1);
	assert_int_equal(granular_ctx->state, RPC_CA_RESTRICTED);

	assert_int_equal(1, rpc_uci_check_granular_is_allowed("radio"));
	assert_int_equal(1, rpc_uci_check_granular_is_allowed("band"));
	assert_int_equal(0, rpc_uci_check_granular_is_allowed("type"));
	assert_int_equal(0, rpc_uci_check_granular_is_allowed("channel"));
	assert_int_equal(0, rpc_uci_check_granular_is_allowed("*"));
	assert_int_equal(0, rpc_uci_check_granular_is_allowed("NON_EXISTENT_OPTION"));
}

static int perform_uci_rename(char *user, char *cfg, char *section, char *option, char *name, char *id)
{
	struct blob_buf bb = {0};
	int rv;

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "ubus_rpc_session", id);
	blobmsg_add_string(&bb, "config", cfg);
	blobmsg_add_string(&bb, "section", section);
	blobmsg_add_string(&bb, "name", name);
	if (option)
		blobmsg_add_string(&bb, "option", option);

	rv = rpc_uci_rename(NULL, NULL, NULL, NULL, bb.head);
	blob_buf_free(&bb);

	return rv;
}

static int perform_uci_delete(char *user, char *cfg, char *section, char *type, char *option, char *options, char *id)
{
	struct blob_buf bb = {0};
	int rv;

	blob_buf_init(&bb, 0);
	if (id)
		blobmsg_add_string(&bb, "ubus_rpc_session", id);
	blobmsg_add_string(&bb, "config", cfg);
	if (section)
		blobmsg_add_string(&bb, "section", section);
	if (type)
		blobmsg_add_string(&bb, "type", type);

	if (option)
		blobmsg_add_string(&bb, "option", option);
	else if (options) {
		void *arr;
		struct json_object *elem, *obj;
		int i;

		obj = json_tokener_parse(options);

		arr = blobmsg_open_array(&bb, "options");
		for (i = 0; i < json_object_array_length(obj); i++) {
			elem = json_object_array_get_idx(obj, i);
			printf("trying to delete val %s\n", json_object_get_string(elem));
			blobmsg_add_string(&bb, "", json_object_get_string(elem));
		}

		blobmsg_close_array(&bb, arr);
		json_object_put(obj);
	}

	rv = rpc_uci_delete(NULL, NULL, NULL, NULL, bb.head);
	blob_buf_free(&bb);

	return rv;
}

static int perform_uci_set(char *user, char *cfg, char *section, char *type, char *values, char *id)
{
	struct blob_buf bb = {0};
	struct json_object *obj;
	void *table;
	int rv;


	obj = json_tokener_parse(values);
	assert_non_null(obj);

	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "config", cfg);
	blobmsg_add_string(&bb, "section", section);
	blobmsg_add_string(&bb, "ubus_rpc_session", id);

	table = blobmsg_open_table(&bb, "values");
	json_object_object_foreach(obj, key, val)
		blobmsg_add_string(&bb, key, json_object_get_string(val));


	blobmsg_close_table(&bb, table);

	rv = rpc_uci_set(NULL, NULL, NULL, NULL, bb.head);

	blob_buf_free(&bb);
	json_object_put(obj);
	return rv;
}

static int perform_uci_add(char *user, char *cfg, char *section, char *type, char *values, char *id)
{
	struct blob_buf bb = {0};
	void *table;
	struct json_object *obj;
	int rv;

	obj = json_tokener_parse(values);

	blob_buf_init(&bb, 0);
	if (id)
		blobmsg_add_string(&bb, "ubus_rpc_session", id);
	blobmsg_add_string(&bb, "config", cfg);
	blobmsg_add_string(&bb, "type", type);
	blobmsg_add_string(&bb, "name", section);

	table = blobmsg_open_table(&bb, "values");
	json_object_object_foreach(obj, key, val) {
		printf("adding key %s val %s\n", key, json_object_get_string(val));
		blobmsg_add_string(&bb, key, json_object_get_string(val));
	}
	blobmsg_close_table(&bb, table);

	rv = rpc_uci_add(NULL, NULL, NULL, NULL, bb.head);

	blob_buf_free(&bb);
	json_object_put(obj);
	return rv;
}

static struct blob_attr *perform_uci_get(char *user, char *cfg, char *section, char *type, char *option, char *id)
{
	struct blob_buf bb = {0};

	/* always reset global ba so failed cases can be verified */
	reset_ba();
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "config", cfg);
	if (id)
		blobmsg_add_string(&bb, "ubus_rpc_session", id);
	if (option)
		blobmsg_add_string(&bb, "option", option);
	if (section)
		blobmsg_add_string(&bb, "section", section);
	if (type)
		blobmsg_add_string(&bb, "type", type);

	rpc_uci_get(NULL, NULL, NULL, NULL, bb.head);

	blob_buf_free(&bb);
	return ba;
}

static void test_granular_access_get_by_config(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct json_object *obj, *tmp;
	struct blob_attr *attr;
	char *str;
	int inner_cnt = 0, outer_cnt = 0;

	attr = perform_uci_get(USER, CFG, NULL, NULL, NULL, e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);
	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_key;

		json_object_object_foreach(outer_val, inner_key, inner_val) {
			(void) inner_val;

			if (inner_key[0] == '.')
				continue;
			inner_cnt++;
		}
		outer_cnt++;
	}

	assert_int_equal(inner_cnt, 4); // 3 because one section lacks a key
	assert_int_equal(3, outer_cnt);

	tmp = json_object_get_by_string(obj, "values.settings.radio");
	assert_string_equal(json_object_get_string(tmp), "wl1");
	tmp = json_object_get_by_string(obj, "values.settings.band");
	assert_string_equal(json_object_get_string(tmp), "b");

	json_object_put(obj);
	free(str);
}

static void test_granular_access_get_by_type(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct json_object *obj, *tmp;
	struct blob_buf bb = {0};
	struct blob_attr *attr;
	char *str;
	int inner_cnt = 0, outer_cnt = 0;

	attr = perform_uci_get(USER, CFG, NULL, "settings", NULL, e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);
	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_key;

		json_object_object_foreach(outer_val, inner_key, inner_val) {
			(void) inner_val;

			if (inner_key[0] == '.')
				continue;
			inner_cnt++;
		}
		outer_cnt++;
	}

	assert_int_equal(inner_cnt, 3);
	assert_int_equal(2, outer_cnt);

	tmp = json_object_get_by_string(obj, "values.settings.radio");
	assert_string_equal(json_object_get_string(tmp), "wl1");
	tmp = json_object_get_by_string(obj, "values.settings.band");
	assert_string_equal(json_object_get_string(tmp), "b");

	json_object_put(obj);
	free(str);
	blob_buf_free(&bb);
}

static void test_granular_access_get_by_option(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct json_object *obj, *tmp;
	struct blob_attr *attr;
	char *str;
	int outer_cnt = 0;

	attr = perform_uci_get(USER, CFG, "settings", "settings", "radio", e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "value");
	assert_non_null(tmp);
	json_object_object_foreach(obj, outer_key, outer_val) {
		(void) outer_key;
		(void) outer_val;
		outer_cnt++;
	}
	assert_int_equal(1, outer_cnt);
	assert_string_equal(json_object_get_string(tmp), "wl1");

	json_object_put(obj);
	free(str);
}

static void test_granular_access_get_by_option_fail(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct json_object *obj, *tmp;
	struct blob_attr *attr;
	char *str;

	attr = perform_uci_get(USER, CFG, "settings", "settings", "type", e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "value");
	assert_null(tmp);

	json_object_put(obj);
	free(str);
}

static void test_uci_granular_parse_match_definition(void **state) {
	int rv;
	struct blob_attr *acl_match_definition = {0};

	rv = uci_granular_parse_match_definition(NULL, NULL);
	assert_int_equal(0, rv);

	rv = uci_granular_parse_match_definition(acl_match_definition, NULL);
	assert_int_equal(0, rv);
}

static void test_granular_access_get_by_type_fail(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct json_object *obj, *tmp;
	struct blob_attr *attr;
	char *str;
	int inner_cnt = 0, outer_cnt = 0;

	attr = perform_uci_get(USER, CFG, NULL, "rule", NULL, e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);

	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_key;

		json_object_object_foreach(outer_val, inner_key, inner_val) {
			(void) inner_val;

			if (inner_key[0] == '.')
				continue;
			inner_cnt++;
		}
		outer_cnt++;
	}

	assert_int_equal(inner_cnt, 0);
	assert_int_equal(0, outer_cnt);

	json_object_put(obj);
	free(str);
}

static void test_granular_multi_session(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct test_env *ex;// = calloc(1, sizeof(struct test_env));
	struct json_object *obj, *tmp;
	struct blob_attr *attr;
	char *str;
	int outer_cnt = 0;

	ex = create_new_session("admin");
	assert_non_null(ex);

	/* test an allowed option */
	attr = perform_uci_get("admin", CFG, "settings", "settings", "type", ex->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "value");
	assert_non_null(tmp);
	json_object_object_foreach(obj, outer_key, outer_val) {
		(void) outer_key;
		(void) outer_val;
		outer_cnt++;
	}
	assert_int_equal(1, outer_cnt);
	assert_string_equal(json_object_get_string(tmp), "broadcom");

	json_object_put(obj);
	free(str);
	outer_cnt = 0;

	/* test a restricted option */
	attr = perform_uci_get("admin", CFG, "settings", "settings", "channel", ex->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "value");
	assert_null(tmp);
	json_object_object_foreach(obj, outer_key1, outer_val2) {
		(void) outer_key1;
		(void) outer_val2;
		outer_cnt++;
	}
	assert_int_equal(0, outer_cnt);

	json_object_put(obj);
	free(str);
	outer_cnt = 0;

	/* test allowed option other user */
	attr = perform_uci_get(USER, CFG, "settings", "settings", "radio", e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "value");
	assert_non_null(tmp);
	json_object_object_foreach(obj, outer_key3, outer_val3) {
		(void) outer_key3;
		(void) outer_val3;
		outer_cnt++;
	}
	assert_int_equal(1, outer_cnt);
	assert_string_equal(json_object_get_string(tmp), "wl1");

	json_object_put(obj);
	free(str);
	outer_cnt = 0;

	/* test restricted option other user */
	attr = perform_uci_get(USER, CFG, "settings", "settings", "channel", e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "value");
	assert_null(tmp);
	json_object_object_foreach(obj, outer_key4, outer_val4) {
		(void) outer_key4;
		(void) outer_val4;
		outer_cnt++;
	}

	assert_int_equal(0, outer_cnt);

	json_object_put(obj);
	free(str);
	outer_cnt = 0;

	/* test an allowed option */
	attr = perform_uci_get("admin", CFG, NULL, "settings", "type", ex->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values.settings.type");
	assert_non_null(tmp);
	json_object_object_foreach(obj, outer_key5, outer_val5) {
		(void) outer_val5;
		if (outer_key5[0] != '.')
			outer_cnt++;
	}
	assert_int_equal(1, outer_cnt);
	assert_string_equal(json_object_get_string(tmp), "broadcom");

	json_object_put(obj);
	free(str);

	destroy_env(ex);
}

static void test_granular_access_get_by_section(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct json_object *obj, *tmp;
	struct blob_attr *attr;
	char *str;
	int outer_cnt = 0;

	attr = perform_uci_get(USER, CFG, "settings", "settings", NULL, e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);
	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_val;

		if (outer_key[0] == '.')
			continue;

		outer_cnt++;
	}

	assert_int_equal(2, outer_cnt);

	tmp = json_object_get_by_string(obj, "values.radio");
	assert_string_equal(json_object_get_string(tmp), "wl1");
	tmp = json_object_get_by_string(obj, "values.band");
	assert_string_equal(json_object_get_string(tmp), "b");

	json_object_put(obj);
	free(str);
}

static void test_granular_access_get_by_section_fail(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct json_object *obj, *tmp;
	struct blob_attr *attr;
	char *str;
	int outer_cnt = 0;

	attr = perform_uci_get(USER, CFG, "fantastic", "rule", NULL, e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);
	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_val;

		if (outer_key[0] == '.')
			continue;

		outer_cnt++;
	}

	assert_int_equal(0, outer_cnt);

	json_object_put(obj);
	free(str);
}


static void test_granular_access_set_by_section(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct json_object *obj, *tmp;
	struct blob_attr *attr;
	char *str;
	struct blob_buf bb = {0};
	int rv;

	/* section with some access */
	rv = perform_uci_set(USER, CFG, "settings", "settings", "{ \"radio\":\"new_radio\" }", e->ses->id);
	assert_int_equal(rv, 0);
	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "settings", "settings", "radio", e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "value");
	assert_non_null(tmp);

	assert_string_equal(json_object_get_string(tmp), "new_radio");
	json_object_put(obj);
	free(str);

	/* section with full access */
	rv = perform_uci_set(USER, CFG, "test", "meta", "{ \"name\":\"test\" }", e->ses->id);
	assert_int_equal(rv, 0);
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "test", "meta", "name", e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "value");
	assert_non_null(tmp);

	assert_string_equal(json_object_get_string(tmp), "test");
	json_object_put(obj);
	free(str);

	/* add by set to section we full access to */
	/*
	perform_uci_set(USER, CFG, "test", "meta", "{ \"garbage\":\"1\" }", e->ses->id);
	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	rpc_uci_revert_commit(NULL, bb.head, true);
	blob_buf_free(&bb);

	attr = perform_uci_get(USER, CFG, "test", "meta", "garbage", NULL);
	assert_non_null(attr);

	tmp = json_object_get_by_string(obj, "value");
	assert_non_null(tmp);

	assert_string_equal(json_object_get_string(tmp), "1");
	json_object_put(obj);
	free(str);
	*/

	blob_buf_free(&bb);
}

static void test_granular_access_set_by_section_fail(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct json_object *obj, *tmp;
	struct blob_attr *attr;
	char *str;
	struct blob_buf bb = {0};


	/* section we do have some access to */
	perform_uci_set(USER, CFG, "settings", "settings", "{ \"type\":\"invalid\" }", e->ses->id);
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "config", "test");
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "settings", "settings", "type", NULL);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "value");
	assert_non_null(tmp);

	assert_string_equal(json_object_get_string(tmp), "broadcom");
	json_object_put(obj);
	free(str);

	/* section we got no access to */
	perform_uci_set(USER, CFG, "fantastic", "rule", "{ \"good\":\"1\" }", e->ses->id);
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "fantastic", "rule", "good", NULL);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "value");
	assert_non_null(tmp);

	assert_string_equal(json_object_get_string(tmp), "0");
	json_object_put(obj);
	free(str);

	/* addition by set to section with no access */
	perform_uci_set(USER, CFG, "settings", "settings", "{ \"garbage\":\"1\" }", e->ses->id);
	rpc_uci_revert_commit(NULL, bb.head, true);
	blob_buf_free(&bb);

	attr = perform_uci_get(USER, CFG, "settings", "settings", "garbage", NULL);
	assert_null(attr);
}

static int teardown(void** state) {
	struct test_env *e = (struct test_env *) *state;
	struct blob_buf bb = {0};
	int rv;

	rv = perform_uci_set(USER, CFG, "settings", "settings", "{ \"radio\":\"wl1\" }", e->ses->id);
	assert_int_equal(rv, 0);
	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	rpc_uci_revert_commit(NULL, bb.head, true);
	blob_buf_free(&bb);
	return 0;
}

static void test_granular_add(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_attr *attr;
	struct blob_buf bb = {0};
	int rv;
	int outer_cnt = 0;
	struct json_object *obj, *tmp;
	char *str;

	/* trying to add on section with full access */
	rv = perform_uci_add(USER, CFG, "new", "meta", "{ \"completely\":\"new\"}", e->ses->id);
	assert_int_equal(rv, 0);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");

	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "new", "meta", NULL, e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);
	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_val;

		if (outer_key[0] == '.')
			continue;

		outer_cnt++;
	}
	assert_int_equal(outer_cnt, 1);

	tmp = json_object_get_by_string(obj, "values.completely");
	assert_string_equal(json_object_get_string(tmp), "new");
	json_object_put(obj);
	free(str);
	outer_cnt = 0;

	/* trying to add on existing section with full access */
	rv = perform_uci_add(USER, CFG, "new", "meta", "{ \"addition\":\"add\", \"addition2\":\"add2\", \"addition3\":\"add3\"}", e->ses->id);
	assert_int_equal(rv, 0);

	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "new", "meta", NULL, e->ses->id);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);
	json_object_object_foreach(tmp, outer_key1, outer_val1) {
		(void) outer_val1;

		if (outer_key1[0] == '.')
			continue;

		outer_cnt++;
	}
	assert_int_equal(outer_cnt, 4);

	tmp = json_object_get_by_string(obj, "values.completely");
	assert_string_equal(json_object_get_string(tmp), "new");
	json_object_put(obj);
	free(str);
	blob_buf_free(&bb);
}

static void test_granular_add_fail(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_attr *attr;
	struct blob_buf bb = {0};
	int rv;
	int outer_cnt = 0;
	struct json_object *obj, *tmp;
	char *str;

	/* trying to add completely new section */
	rv = perform_uci_add(USER, CFG, "new_section", "new_type", "{ \"completley\":\"new\"}", e->ses->id);
	assert_int_equal(rv, 6);

	/* trying to add on top of an existing section on option we have access to */
	rv = perform_uci_add(USER, CFG, "settings", "settings", "{ \"radio\":\"fake\"}", e->ses->id);
	assert_int_equal(rv, 6);

	/* trying to add on top of an existing section on option with no access */
	rv = perform_uci_add(USER, CFG, "settings", "settings", "{ \"test\":\"fake\"}", e->ses->id);
	assert_int_equal(rv, 6);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");

	rpc_uci_revert_commit(NULL, bb.head, true);

	blob_buf_free(&bb);

	attr = perform_uci_get(USER, CFG, "new_section", NULL, NULL, NULL);
	assert_null(attr);

	attr = perform_uci_get(USER, CFG, "settings", "settings", NULL, NULL);
	assert_non_null(attr);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);
	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_val;

		if (outer_key[0] == '.')
			continue;

		outer_cnt++;
	}
	assert_int_equal(outer_cnt, 5);

	tmp = json_object_get_by_string(obj, "values.radio");
	assert_string_equal(json_object_get_string(tmp), "wl1");
	json_object_put(obj);
	free(str);
}

static void test_granular_delete_by_section_fail(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_attr *attr;
	struct blob_buf bb = {0};
	int rv;
	int outer_cnt = 0;
	struct json_object *obj, *tmp;
	char *str;

	/* delete options of section we have limited access to */
	rv = perform_uci_delete(USER, CFG, "settings", "settings", NULL, "[\"radio\", \"type\"]", e->ses->id);
	assert_int_equal(rv, 6);

	rv = perform_uci_delete(USER, CFG, "settings", "settings", "radio", NULL, e->ses->id);
	assert_int_equal(rv, 6);

	rv = perform_uci_delete(USER, CFG, "settings", "settings", NULL, NULL, e->ses->id);
	assert_int_equal(rv, 6);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	blobmsg_add_string(&bb, "ubus_rpc_session", e->ses->id);
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "settings", "settings", NULL, NULL);
	assert_non_null(attr);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);
	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_val;

		if (outer_key[0] == '.')
			continue;

		outer_cnt++;
	}
	assert_int_equal(outer_cnt, 5);

	tmp = json_object_get_by_string(obj, "values.radio");
	assert_non_null(tmp);

	json_object_put(obj);
	free(str);
	blob_buf_free(&bb);
}

static void test_granular_delete_by_type_fail(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_attr *attr;
	struct blob_buf bb = {0};
	int rv;
	int inner_cnt = 0, outer_cnt = 0;
	struct json_object *obj, *tmp;
	char *str;

	/* delete options of section we have limited access to */
	rv = perform_uci_delete(USER, CFG, NULL, "settings", NULL, "[\"radio\", \"type\"]", e->ses->id);
	assert_int_equal(rv, 6);

	rv = perform_uci_delete(USER, CFG, NULL, "settings", "radio", NULL, e->ses->id);
	assert_int_equal(rv, 6);

	rv = perform_uci_delete(USER, CFG, NULL, "settings", NULL, NULL, e->ses->id);
	assert_int_equal(rv, 6);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	blobmsg_add_string(&bb, "ubus_rpc_session", e->ses->id);
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, NULL, "settings", NULL, NULL);
	assert_non_null(attr);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_key;

		json_object_object_foreach(outer_val, inner_key, inner_val) {
			(void) inner_val;

			if (inner_key[0] == '.')
				continue;
			inner_cnt++;
		}

		outer_cnt++;
	}
	assert_int_equal(outer_cnt, 2);
	assert_int_equal(inner_cnt, 7);

	tmp = json_object_get_by_string(obj, "values.settings.radio");
	assert_non_null(tmp);

	json_object_put(obj);
	free(str);
	blob_buf_free(&bb);
}

static void test_granular_delete_by_section(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_attr *attr;
	struct blob_buf bb = {0};
	int rv;
	int outer_cnt = 0;
	struct json_object *obj, *tmp;
	char *str;

	/* delete options of section we have full access to */
	rv = perform_uci_delete(USER, CFG, "new", "meta", NULL, "[\"addition3\", \"addition2\"]", e->ses->id);
	assert_int_equal(rv, 0);

	/* delete option of section we have full access to */
	rv = perform_uci_delete(USER, CFG, "new", "meta", "addition", NULL, e->ses->id);
	assert_int_equal(rv, 0);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	blobmsg_add_string(&bb, "ubus_rpc_session", e->ses->id);
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "new", "meta", NULL, NULL);
	assert_non_null(attr);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);
	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_val;

		if (outer_key[0] == '.')
			continue;

		outer_cnt++;
	}
	assert_int_equal(outer_cnt, 1);

	json_object_put(obj);
	free(str);

	/* delete entire section */
	rv = perform_uci_delete(USER, CFG, "new", "meta", NULL, NULL, e->ses->id);
	assert_int_equal(rv, 0);

	rv = rpc_uci_revert_commit(NULL, bb.head, true);
	assert_int_equal(rv, 0);

	attr = perform_uci_get(USER, CFG, "new", "meta", NULL, NULL);
	assert_null(attr);

	blob_buf_free(&bb);
}

static void test_granular_delete_by_type(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_attr *attr;
	struct blob_buf bb = {0};
	int rv;
	int outer_cnt = 0;
	struct json_object *obj, *tmp;
	char *str;

	/* delete options of type we have full access to */
	rv = perform_uci_delete(USER, CFG, NULL, "meta", NULL, "[\"addition3\", \"addition2\"]", e->ses->id);
	assert_int_equal(rv, 0);

	/* delete option of type we have full access to */
	rv = perform_uci_delete(USER, CFG, NULL, "meta", "addition", NULL, e->ses->id);
	assert_int_equal(rv, 0);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	blobmsg_add_string(&bb, "ubus_rpc_session", e->ses->id);
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, NULL, "meta", NULL, NULL);
	assert_non_null(attr);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	assert_non_null(tmp);
	json_object_object_foreach(tmp, outer_key, outer_val) {
		(void) outer_val;

		if (outer_key[0] == '.')
			continue;

		outer_cnt++;
	}
	assert_int_equal(outer_cnt, 2);

	json_object_put(obj);
	free(str);
	outer_cnt = 0;

	/* delete all sections of type */
	rv = perform_uci_delete(USER, CFG, NULL, "meta", NULL, NULL, e->ses->id);
	assert_int_equal(rv, 0);

	rpc_uci_revert_commit(NULL, bb.head, true);

	/* slight behavior difference between getting empty option, section and type? */
	attr = perform_uci_get(USER, CFG, NULL, "meta", NULL, NULL);

	str = blobmsg_format_json(attr, true);
	assert_non_null(str);

	obj = json_tokener_parse(str);
	assert_non_null(obj);

	printf("%s %d: str=%s\n", __func__, __LINE__, str);
	printf("%s %d: obj=%s\n", __func__, __LINE__, json_object_get_string(obj));

	tmp = json_object_get_by_string(obj, "values");
	json_object_object_foreach(tmp, outer_key1, outer_val1) {
		(void) outer_val1;

		if (outer_key1[0] == '.')
			continue;

		outer_cnt++;
	}

	assert_int_equal(outer_cnt, 0);

	json_object_put(obj);
	free(str);
	blob_buf_free(&bb);
}

static void test_granular_rename_section(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_attr *attr;
	struct blob_buf bb = {0};
	int rv;

	/* rename section we have full access to */
	rv = perform_uci_rename(USER, CFG, "test", NULL, "newer", e->ses->id);
	assert_int_equal(rv, 0);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	blobmsg_add_string(&bb, "ubus_rpc_session", e->ses->id);
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "test", "meta", NULL, NULL);
	assert_null(attr);

	attr = perform_uci_get(USER, CFG, "newer", "meta", NULL, NULL);
	assert_non_null(attr);

	blob_buf_free(&bb);
}

static void test_granular_rename_option(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_attr *attr;
	struct blob_buf bb = {0};
	int rv;

	/* rename option in section we have full access to */
	rv = perform_uci_rename(USER, CFG, "test", "name", "rename", e->ses->id);
	assert_int_equal(rv, 0);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	blobmsg_add_string(&bb, "ubus_rpc_session", e->ses->id);
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "test", "meta", "name", NULL);
	assert_null(attr);

	attr = perform_uci_get(USER, CFG, "test", "meta", "rename", NULL);
	assert_non_null(attr);

	blob_buf_free(&bb);
}

static void test_granular_rename_option_fail(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_attr *attr;
	struct blob_buf bb = {0};
	int rv;

	/* rename option in section with limited access */
	rv = perform_uci_rename(USER, CFG, "settings", "radio", "new_radio", e->ses->id);
	assert_int_equal(rv, 6);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	blobmsg_add_string(&bb, "ubus_rpc_session", e->ses->id);
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "settings", "settings", "new_radio", NULL);
	assert_null(attr);

	attr = perform_uci_get(USER, CFG, "settings", "settings", "radio", NULL);
	assert_non_null(attr);

	blob_buf_free(&bb);
}

static void test_granular_rename_section_fail(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_attr *attr;
	struct blob_buf bb = {0};
	int rv;

	/* rename section we have full access to */
	rv = perform_uci_rename(USER, CFG, "settings", NULL, "new_settings", e->ses->id);
	assert_int_equal(rv, 6);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	blobmsg_add_string(&bb, "ubus_rpc_session", e->ses->id);
	rpc_uci_revert_commit(NULL, bb.head, true);

	attr = perform_uci_get(USER, CFG, "new_settings", "settings", NULL, NULL);
	assert_null(attr);

	attr = perform_uci_get(USER, CFG, "settings", "settings", NULL, NULL);
	assert_non_null(attr);

	blob_buf_free(&bb);
}

static int setup(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_buf bb = {0};
	int rv;

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	rv = perform_uci_add(USER, CFG, "new", "meta", "{ \"completely\":\"new\", \"addition\":\"add\", \"addition2\":\"add2\", \"addition3\":\"add3\"}", e->ses->id);
	assert_int_equal(rv, 0);
	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	blobmsg_add_string(&bb, "ubus_rpc_session", e->ses->id);
	rv = rpc_uci_revert_commit(NULL, bb.head, true);
	assert_int_equal(rv, 0);
	blob_buf_free(&bb);
	return 0;
}

static int setup_rename(void **state)
{
	struct test_env *e = (struct test_env *) *state;
	struct blob_buf bb = {0};
	int rv;

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "config", "test");
	rv = perform_uci_add(USER, CFG, "test", "meta", "{ \"name\":\"rpcd_granular\"}", e->ses->id);
	assert_int_equal(rv, 0);
	rpc_uci_revert_commit(NULL, bb.head, true);

	blob_buf_free(&bb);
	return 0;
}

/**
 * This is run once before all group tests
 */
static int group_setup(void** state) {
	struct test_env *e;// = calloc(1, sizeof(struct test_env));

	remove("/etc/config/test");
	cp("/etc/config/test", "files/etc/config/test");

	rpc_uci_init_uci();
	rpc_init_default_session();

	e = create_new_session("user");
	if (!e)
		return -1;

	printf("session: %s\n", e->ses->id);
	*state = e;
	return 0;
}

static int group_teardown(void** state)
{
	struct test_env *e = (struct test_env *) *state;

	rpc_uci_granular_destroy();

	rpc_uci_clean();

	destroy_env(e);

	rpc_session_destroy(rpc_session_get("00000000000000000000000000000000"));

	return 0;
}

int main(void) {
	const struct CMUnitTest tests[] = {
		cmocka_unit_test(test_settings_name_access),
		cmocka_unit_test(test_granular_access_get_by_type),
		cmocka_unit_test(test_granular_access_get_by_type_fail),
		cmocka_unit_test(test_uci_granular_parse_match_definition),
		cmocka_unit_test(test_granular_access_get_by_config),
		cmocka_unit_test(test_granular_access_get_by_option),
		cmocka_unit_test(test_granular_access_get_by_option_fail),
		cmocka_unit_test(test_granular_multi_session),
		cmocka_unit_test(test_granular_access_get_by_section),
		cmocka_unit_test(test_granular_access_get_by_section_fail),
		cmocka_unit_test_setup_teardown(test_granular_access_set_by_section, NULL, teardown),
		cmocka_unit_test(test_granular_access_set_by_section_fail),
		cmocka_unit_test(test_granular_add), // some form of memory leak?
		cmocka_unit_test(test_granular_add_fail), // some form of memory leak?
		cmocka_unit_test_setup_teardown(test_granular_delete_by_section, setup, NULL), // some form of memory leak?
		cmocka_unit_test(test_granular_delete_by_section_fail),
		cmocka_unit_test_setup_teardown(test_granular_delete_by_type, setup, NULL), // some form of leak??
		cmocka_unit_test(test_granular_delete_by_type_fail),
		cmocka_unit_test_setup_teardown(test_granular_rename_option, setup_rename, NULL), // some form of leak?
		cmocka_unit_test_setup_teardown(test_granular_rename_section, setup_rename, NULL), // some form of leak
		cmocka_unit_test(test_granular_rename_option_fail),
		cmocka_unit_test(test_granular_rename_section_fail),
	};

	return cmocka_run_group_tests(tests, group_setup, group_teardown);
}
